@@ -201,13 +201,13 @@ int set_infra_mode(struct wrap_ndis_device *wnd,
 	if (wnd->infrastructure_mode == mode)
 		TRACEEXIT2(return 0);
 
+	memset(&wnd->essid, 0, sizeof(wnd->essid));
 	res = miniport_set_int(wnd, OID_802_11_INFRASTRUCTURE_MODE, mode);
 	if (res) {
 		WARNING("setting operating mode to %d failed (%08X)",
 			mode, res);
 		TRACEEXIT2(return -EINVAL);
 	}
-
 	/* NDIS drivers clear keys when infrastructure mode is
 	 * changed. But Linux tools assume otherwise. So set the
 	 * keys */
@@ -216,7 +216,6 @@ int set_infra_mode(struct wrap_ndis_device *wnd,
 			add_wep_key(wnd, wnd->encr_info.keys[i].key,
 				    wnd->encr_info.keys[i].length, i);
 	}
-	wnd->infrastructure_mode = mode;
 	TRACEEXIT2(return 0);
 }
 
@@ -968,10 +967,10 @@ static int iw_set_nick(struct net_device *dev, struct iw_request_info *info,
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
 
-	if (wrqu->data.length > IW_ESSID_MAX_SIZE)
+	if (wrqu->data.length > IW_ESSID_MAX_SIZE || wrqu->data.length <= 0)
 		return -EINVAL;
 	memcpy(wnd->nick, extra, wrqu->data.length);
-	wnd->nick[IW_ESSID_MAX_SIZE] = 0;
+	wnd->nick[wrqu->data.length-1] = 0;
 	return 0;
 }
 
@@ -980,8 +979,8 @@ static int iw_get_nick(struct net_device *dev, struct iw_request_info *info,
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
 
-	memcpy(extra, wnd->nick, IW_ESSID_MAX_SIZE+1);
 	wrqu->data.length = strlen(wnd->nick);
+	memcpy(extra, wnd->nick, wrqu->data.length);
 	return 0;
 }
 
