@@ -23,7 +23,7 @@
 #define MAX_ALLOCATED_NDIS_BUFFERS 20
 
 static workqueue_struct_t *ndis_wq;
-static void ndis_worker(void *dummy);
+static void ndis_worker(worker_param_t dummy);
 static work_struct_t ndis_work;
 static struct nt_list ndis_worker_list;
 static NT_SPIN_LOCK ndis_work_list_lock;
@@ -139,9 +139,8 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMRegisterMiniport,3)
 			"tx_data", "return_packet", "send_packets",
 			"alloc_complete", "co_create_vc", "co_delete_vc",
 			"co_activate_vc", "co_deactivate_vc",
-			"co_send_packets", "co_request",
-			"cancel_send_packets", "pnp_event_notify",
-			"shutdown",
+			"co_send_packets", "co_request", "cancel_send_packets",
+			"pnp_event_notify", "shutdown",
 		};
 		func = (void **)&ndis_driver->miniport.query;
 		for (i = 0; i < (sizeof(miniport_funcs) /
@@ -371,7 +370,7 @@ wstdcall ULONG WIN_FUNC(NDIS_BUFFER_TO_SPAN_PAGES,1)
 #ifdef VT6655
 	/* VIA VT6655 works with this bogus computation, but not with
 	 * correct computation with SPAN_PAGES */
-	if (1) {
+	do {
 		ULONG_PTR start, end;
 		unsigned long ptr;
 
@@ -379,7 +378,7 @@ wstdcall ULONG WIN_FUNC(NDIS_BUFFER_TO_SPAN_PAGES,1)
 		start = ptr & (PAGE_SIZE - 1);
 		end = (ptr + length + PAGE_SIZE - 1) & PAGE_MASK;
 		n = (end - start) / PAGE_SIZE;
-	}
+	} while (0);
 #else
 	n = SPAN_PAGES(MmGetMdlVirtualAddress(buffer), length);
 #endif
@@ -667,12 +666,15 @@ wstdcall ULONG WIN_FUNC(NdisReadPciSlotInformation,5)
 {
 	struct wrap_device *wd = nmb->wnd->wd;
 	ULONG i;
-	TRACEENTER4("%d", len);
 	for (i = 0; i < len; i++)
 		if (pci_read_config_byte(wd->pci.pdev, offset + i, &buf[i]) !=
 		    PCIBIOS_SUCCESSFUL)
 			break;
-	TRACEEXIT4(return i);
+	DBG_BLOCK(2) {
+		if (i != len)
+			WARNING("%u, %u", i, len);
+	}
+	return i;
 }
 
 wstdcall ULONG WIN_FUNC(NdisImmediateReadPciSlotInformation,5)
@@ -688,12 +690,15 @@ wstdcall ULONG WIN_FUNC(NdisWritePciSlotInformation,5)
 {
 	struct wrap_device *wd = nmb->wnd->wd;
 	ULONG i;
-	TRACEENTER4("%d", len);
 	for (i = 0; i < len; i++)
 		if (pci_write_config_byte(wd->pci.pdev, offset + i, buf[i]) !=
 		    PCIBIOS_SUCCESSFUL)
 			break;
-	TRACEEXIT4(return i);
+	DBG_BLOCK(2) {
+		if (i != len)
+			WARNING("%u, %u", i, len);
+	}
+	return i;
 }
 
 wstdcall void WIN_FUNC(NdisReadPortUchar,3)
@@ -872,7 +877,7 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMAllocateMapRegisters,5)
 		WARNING("Windows driver %s requesting too many (%u) "
 			"map registers", wnd->wd->driver->name, basemap);
 		/* As per NDIS, NDIS_STATUS_RESOURCES should be
-		 * retrned, but with that Atheros PCI driver fails -
+		 * returned, but with that Atheros PCI driver fails -
 		 * for now tolerate it */
 //		TRACEEXIT2(return NDIS_STATUS_RESOURCES);
 	}
@@ -1101,8 +1106,6 @@ wstdcall void WIN_FUNC(NdisAllocateBuffer,5)
 	if (descr) {
 		typeof(descr->flags) flags = descr->flags;
 		MmInitializeMdl(descr, virt, length);
-//		MmBuildMdlForNonPagedPool(descr);
-		descr->flags |= MDL_SOURCE_IS_NONPAGED_POOL;
 		if (flags & MDL_CACHE_ALLOCATED)
 			descr->flags |= MDL_CACHE_ALLOCATED;
 	} else {
@@ -1115,8 +1118,8 @@ wstdcall void WIN_FUNC(NdisAllocateBuffer,5)
 		DBGTRACE4("allocated buffer %p for %p, %d",
 			  descr, virt, length);
 		atomic_inc_var(pool->num_allocated_descr);
-		MmBuildMdlForNonPagedPool(descr);
 	}
+	MmBuildMdlForNonPagedPool(descr);
 //	descr->flags |= MDL_ALLOCATED_FIXED_SIZE |
 //		MDL_MAPPED_TO_SYSTEM_VA | MDL_PAGES_LOCKED;
 	descr->pool = pool;
@@ -1145,7 +1148,7 @@ wstdcall void WIN_FUNC(NdisFreeBuffer,1)
 		descr->pool = NULL;
 		free_mdl(descr);
 	} else
-		atomic_insert_list_head(pool->free_descr, descr->next, descr);
+		atomic_insert_list_head(descr->next, pool->free_descr, descr);
 	TRACEEXIT4(return);
 }
 
@@ -1188,8 +1191,8 @@ wstdcall void WIN_FUNC(NdisQueryBuffer,3)
 	if (virt)
 		*virt = MmGetSystemAddressForMdl(buffer);
 	*length = MmGetMdlByteCount(buffer);
-	DBGTRACE4("%u", *length);
-	TRACEEXIT4(return);
+	DBGTRACE4("%p, %u", virt? *virt : NULL, *length);
+	return;
 }
 
 wstdcall void WIN_FUNC(NdisQueryBufferSafe,4)
@@ -1200,7 +1203,7 @@ wstdcall void WIN_FUNC(NdisQueryBufferSafe,4)
 	if (virt)
 		*virt = MmGetSystemAddressForMdlSafe(buffer, priority);
 	*length = MmGetMdlByteCount(buffer);
-	DBGTRACE4("%u", *length);
+	DBGTRACE4("%p, %u", virt? *virt : NULL, *length);
 }
 
 wstdcall void *WIN_FUNC(NdisBufferVirtualAddress,1)
@@ -1416,7 +1419,7 @@ wstdcall void WIN_FUNC(NdisAllocatePacket,3)
 	ndis_packet->private.oob_offset = packet_length -
 		sizeof(struct ndis_packet_oob_data);
 	ndis_packet->private.packet_flags = fPACKET_ALLOCATED_BY_NDIS;
-//		| NDIS_PROTOCOL_ID_TCP_IP;
+	ndis_packet->private.flags = NDIS_PROTOCOL_ID_TCP_IP;
 	ndis_packet->private.pool = pool;
 	*packet = ndis_packet;
 	*status = NDIS_STATUS_SUCCESS;
@@ -1449,7 +1452,7 @@ wstdcall void WIN_FUNC(NdisFreePacket,1)
 	} else {
 		struct ndis_packet_oob_data *oob_data;
 		oob_data = NDIS_PACKET_OOB_DATA(descr);
-		atomic_insert_list_head(pool->free_descr, oob_data->next, descr);
+		atomic_insert_list_head(oob_data->next, pool->free_descr, descr);
 	}
 	TRACEEXIT4(return);
 }
@@ -1609,7 +1612,7 @@ wstdcall void wrap_miniport_timer(struct kdpc *kdpc, void *ctx, void *arg1,
 	/* already called at DISPATCH_LEVEL */
 	if (!deserialized_driver(nmb->wnd))
 		serialize_lock(nmb->wnd);
-	LIN2WIN4(timer->func, kdpc, timer->ctx, kdpc->arg1, kdpc->arg2);
+	LIN2WIN4(timer->func, NULL, timer->ctx, NULL, NULL);
 	if (!deserialized_driver(nmb->wnd))
 		serialize_unlock(nmb->wnd);
 	TRACEEXIT5(return);
@@ -1692,7 +1695,7 @@ wstdcall void WIN_FUNC(NdisReadNetworkAddress,4)
 	int ret;
 
 	TRACEENTER1("");
-	RtlInitAnsiString(&ansi, "NetworkAddress");
+	RtlInitAnsiString(&ansi, "mac_address");
 	*len = 0;
 	*status = NDIS_STATUS_FAILURE;
 	if (RtlAnsiStringToUnicodeString(&key, &ansi, TRUE) != STATUS_SUCCESS)
@@ -1708,9 +1711,9 @@ wstdcall void WIN_FUNC(NdisReadNetworkAddress,4)
 		if (ret != NDIS_STATUS_SUCCESS)
 			TRACEEXIT1(return);
 
-		ret = sscanf(ansi.buf, MACSTR, MACINTADR(int_mac));
+		ret = sscanf(ansi.buf, MACSTRSEP, MACINTADR(int_mac));
 		if (ret != ETH_ALEN)
-			ret = sscanf(ansi.buf, MACSTRSEP, MACINTADR(int_mac));
+			ret = sscanf(ansi.buf, MACSTR, MACINTADR(int_mac));
 		RtlFreeAnsiString(&ansi);
 		if (ret == ETH_ALEN) {
 			int i;
@@ -1795,7 +1798,7 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMRegisterInterrupt,7)
 		    ndis_irq, vector, level, req_isr, shared, mode);
 
 	ndis_irq->irq.irq = vector;
-	ndis_irq->wnd = wnd;
+	ndis_irq->nmb = nmb;
 	ndis_irq->req_isr = req_isr;
 	if (shared && !req_isr)
 		WARNING("shared but dynamic interrupt!");
@@ -1817,21 +1820,21 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMRegisterInterrupt,7)
 wstdcall void WIN_FUNC(NdisMDeregisterInterrupt,1)
 	(struct ndis_irq *ndis_irq)
 {
-	struct wrap_ndis_device *wnd;
+	struct ndis_miniport_block *nmb;
 
 	TRACEENTER1("%p", ndis_irq);
 
 	if (!ndis_irq)
 		TRACEEXIT1(return);
-	wnd = ndis_irq->wnd;
-	if (!wnd)
+	nmb = ndis_irq->nmb;
+	if (!nmb)
 		TRACEEXIT1(return);
 
-	free_irq(ndis_irq->irq.irq, wnd);
-	tasklet_kill(&wnd->irq_tasklet);
+	free_irq(ndis_irq->irq.irq, nmb->wnd);
+	tasklet_kill(&nmb->wnd->irq_tasklet);
 	ndis_irq->enabled = 0;
-	ndis_irq->wnd = NULL;
-	wnd->ndis_irq = NULL;
+	ndis_irq->nmb = NULL;
+	nmb->wnd->ndis_irq = NULL;
 	TRACEEXIT1(return);
 }
 
@@ -1864,11 +1867,15 @@ wstdcall void WIN_FUNC(NdisMIndicateStatus,4)
 	TRACEENTER2("status=0x%x len=%d", status, len);
 	switch (status) {
 	case NDIS_STATUS_MEDIA_DISCONNECT:
+		if (!netif_carrier_ok(wnd->net_dev))
+			break;
 		netif_carrier_off(wnd->net_dev);
 		set_bit(LINK_STATUS_CHANGED, &wnd->wrap_ndis_pending_work);
 		schedule_wrap_work(&wnd->wrap_ndis_work);
 		break;
 	case NDIS_STATUS_MEDIA_CONNECT:
+		if (netif_carrier_ok(wnd->net_dev))
+			break;
 		netif_carrier_on(wnd->net_dev);
 		set_bit(LINK_STATUS_CHANGED, &wnd->wrap_ndis_pending_work);
 		schedule_wrap_work(&wnd->wrap_ndis_work);
@@ -1878,7 +1885,6 @@ wstdcall void WIN_FUNC(NdisMIndicateStatus,4)
 			break;
 		si = buf;
 		DBGTRACE2("status_type=%d", si->status_type);
-
 		switch (si->status_type) {
 		case Ndis802_11StatusType_Authentication:
 			buf = (char *)buf + sizeof(*si);
@@ -2033,6 +2039,8 @@ wstdcall void NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 						 NormalPagePriority);
 		DBGTRACE3("%d, %d", length, total_length);
 		oob_data = NDIS_PACKET_OOB_DATA(packet);
+		DBGTRACE3("0x%x, 0x%x, %Lu", packet->private.flags,
+			  packet->private.packet_flags, oob_data->time_rxed);
 		skb = dev_alloc_skb(total_length);
 		if (skb) {
 			while (buffer) {
@@ -2317,13 +2325,13 @@ wstdcall void WIN_FUNC(NdisMSleep,1)
 	delay = USEC_TO_HZ(us);
 	sleep_hz(delay);
 	DBGTRACE4("%p: done", current);
-	TRACEEXIT4(return);
 }
 
 wstdcall void WIN_FUNC(NdisGetCurrentSystemTime,1)
 	(LARGE_INTEGER *time)
 {
 	*time = ticks_1601();
+	DBGTRACE5("%Lu, %lu", *time, jiffies);
 }
 
 wstdcall NDIS_STATUS WIN_FUNC(NdisMRegisterIoPortRange,4)
@@ -2452,7 +2460,7 @@ wstdcall void NdisMResetComplete(struct ndis_miniport_block *nmb,
 	TRACEEXIT3(return);
 }
 
-static void ndis_worker(void *dummy)
+static void ndis_worker(worker_param_t dummy)
 {
 	KIRQL irql;
 	struct ndis_work_entry *ndis_work_entry;
@@ -2585,6 +2593,24 @@ wstdcall void WIN_FUNC(NdisMCoRequestComplete,3)
 	TRACEEXIT3(return);
 }
 
+wstdcall NDIS_STATUS WIN_FUNC(NdisIMNotifiyPnPEvent,2)
+	(struct ndis_miniport_block *nmb, struct net_pnp_event *event)
+{
+	TRACEENTER2("%p, %d", nmb, event->code);
+	/* NdisWrapper never calls protocol's pnp event notifier, so
+	 * nothing to do here */
+	TRACEEXIT2(return NDIS_STATUS_SUCCESS);
+}
+
+wstdcall void WIN_FUNC(NdisCompletePnPEvent,2)
+	(NDIS_STATUS status, void *handle, struct net_pnp_event *event)
+{
+	TRACEENTER2("%d, %p, %d", status, handle, event->code);
+	/* NdisWrapper never calls protocol's pnp event notifier, so
+	 * nothing to do here */
+	TRACEEXIT2(return);
+}
+
 wstdcall NDIS_STATUS WIN_FUNC(NdisMSetMiniportSecondary,2)
 	(struct ndis_miniport_block *nmb2, struct ndis_miniport_block *nmb1)
 {
