@@ -83,6 +83,23 @@ struct ndis_buffer_pool {
 
 #define PROTOCOL_RESERVED_SIZE_IN_PACKET (4 * sizeof(void *))
 
+struct transport_header_offset {
+	USHORT protocol_type;
+	USHORT header_offset;
+};
+
+struct ndis_network_address {
+	USHORT length;
+	USHORT type;
+	UCHAR address[1];
+};
+
+struct ndis_network_address_list {
+	LONG count;
+	USHORT type;
+	struct ndis_network_address address[1];
+};
+
 struct ndis_tcp_ip_checksum_packet_info {
 	union {
 		struct {
@@ -200,7 +217,7 @@ struct ndis_packet_oob_data {
 	union {
 		ULONGLONG time_to_tx;
 		ULONGLONG time_txed;
-	} oob_tx;
+	};
 	ULONGLONG time_rxed;
 	UINT header_size;
 	UINT mediaspecific_size;
@@ -335,6 +352,22 @@ struct ndis_pm_wakeup_capabilities {
 #define NDIS_PNP_WAKE_UP_PATTERN_MATCH			0x00000002
 #define NDIS_PNP_WAKE_UP_LINK_CHANGE			0x00000004
 
+enum net_pnp_event_code {
+	NetEventSetPower, NetEventQueryPower, NetEventQueryRemoveDevice,
+	NetEventCancelRemoveDevice, NetEventReconfigure, NetEventBindList,
+	NetEventBindsComplete, NetEventPnPCapabilities, NetEventMaximum
+};
+
+struct net_pnp_event {
+	enum net_pnp_event_code code;
+	void *buf;
+	ULONG buf_length;
+	ULONG_PTR ndis_reserved[4];
+	ULONG_PTR transport_reserved[4];
+	ULONG_PTR tdi_reserved[4];
+	ULONG_PTR tdi_client_reserved[4];
+};
+
 struct ndis_pnp_capabilities {
 	ULONG flags;
 	struct ndis_pm_wakeup_capabilities wakeup_capa;
@@ -441,6 +474,8 @@ struct ndis_work_entry {
 	struct ndis_work_item *ndis_work_item;
 };
 
+struct ndis_miniport_block;
+
 struct ndis_irq {
 	/* void *intr_obj is used for irq */
 	union {
@@ -453,7 +488,7 @@ struct ndis_irq {
 	ndis_isr_handler isr;
 	void *dpc;
 	struct kdpc intr_dpc;
-	struct wrap_ndis_device *wnd;
+	struct ndis_miniport_block *nmb;
 	UCHAR dpc_count;
 	/* unsigned char filler1 is used for enabled */
 	UCHAR enabled;
@@ -519,8 +554,9 @@ struct ndis_wireless_stats {
 
 enum ndis_status_type {
 	Ndis802_11StatusType_Authentication,
-	Ndis802_11StatusType_PMKID_CandidateList = 2,
-	Ndis802_11StatusType_MediaStreamMode, Ndis802_11StatusType_RadioState,
+	Ndis802_11StatusType_MediaStreamMode,
+	Ndis802_11StatusType_PMKID_CandidateList,
+	Ndis802_11StatusType_RadioState,
 };
 
 struct ndis_status_indication {
@@ -620,7 +656,7 @@ struct ndis_miniport_timer {
 	struct kdpc kdpc;
 	DPC func;
 	void *ctx;
-	void *nmb;
+	struct ndis_miniport_block *nmb;
 	struct ndis_miniport_timer *next;
 };
 
@@ -641,8 +677,6 @@ struct ndis_reference {
 	BOOLEAN closing;
 };
 
-struct ndis_miniport_block;
-
 struct ndis_miniport_interrupt {
 	void *object;
 	NT_SPIN_LOCK dpc_count_lock;
@@ -650,7 +684,7 @@ struct ndis_miniport_interrupt {
 	ndis_isr_handler irq_th;
 	ndis_interrupt_handler irq_bh;
 	struct kdpc interrupt_dpc;
-	struct ndis_miniport_block *miniport;
+	struct ndis_miniport_block *nmb;
 	UCHAR dpc_count;
 	BOOLEAN filler1;
 	struct nt_event dpcs_completed_event;
@@ -822,7 +856,7 @@ struct wrap_ndis_device {
 	struct timer_list stats_timer;
 	unsigned long scan_timestamp;
 	struct encr_info encr_info;
-	char nick[IW_ESSID_MAX_SIZE+1];
+	char nick[IW_ESSID_MAX_SIZE];
 	struct ndis_essid essid;
 	struct auth_encr_capa capa;
 	enum authentication_mode auth_mode;
@@ -848,9 +882,9 @@ struct wrap_ndis_device {
 	struct ndis_tcp_ip_checksum_packet_info tx_csum_info;
 	enum ndis_physical_medium physical_medium;
 	u32 ndis_wolopts;
-	BOOLEAN pm_capa;
 	struct nt_list timer_list;
 	char netdev_name[IFNAMSIZ];
+	int drv_ndis_version;
 };
 
 struct ndis_pmkid_candidate {
@@ -1209,11 +1243,20 @@ void NdisReadConfiguration(NDIS_STATUS *status,
 #define NDIS_ATTRIBUTE_NOT_CO_NDIS              0x00000100
 #define NDIS_ATTRIBUTE_USES_SAFE_BUFFER_APIS    0x00000200
 
-#define NDIS_FLAGS_PROTOCOL_ID_MASK		0x0000000F
-#define NDIS_PROTOCOL_ID_TCP_IP			0x02
-
 #define OID_TCP_TASK_OFFLOAD			0xFC010201
 
+#define NDIS_MAC_OPTION_COPY_LOOKAHEAD_DATA	0x00000001
+#define NDIS_MAC_OPTION_RECEIVE_SERIALIZED	0x00000002
+#define NDIS_MAC_OPTION_TRANSFERS_NOT_PEND	0x00000004
+#define NDIS_MAC_OPTION_NO_LOOPBACK		0x00000008
+#define NDIS_MAC_OPTION_FULL_DUPLEX		0x00000010
+#define NDIS_MAC_OPTION_EOTX_INDICATION		0x00000020
+#define NDIS_MAC_OPTION_8021P_PRIORITY		0x00000040
+#define NDIS_MAC_OPTION_SUPPORTS_MAC_ADDRESS_OVERWRITE	0x00000080
+#define NDIS_MAC_OPTION_RECEIVE_AT_DPC		0x00000100
+#define NDIS_MAC_OPTION_8021Q_VLAN		0x00000200
+#define NDIS_MAC_OPTION_RESERVED		0x80000000
+
 #define deserialized_driver(wnd) (wnd->attributes & NDIS_ATTRIBUTE_DESERIALIZE)
 
 static inline void serialize_lock(struct wrap_ndis_device *wnd)
