@@ -50,7 +50,7 @@ static NT_SPIN_LOCK dispatcher_lock;
 static struct thread_event_waitq *thread_event_waitq_pool;
 
 NT_SPIN_LOCK ntoskernel_lock;
-static kmem_cache_t *mdl_cache;
+static void *mdl_cache;
 static struct nt_list wrap_mdl_list;
 
 /* use tasklet instead worker to execute kdpc's */
@@ -61,7 +61,7 @@ static struct tasklet_struct kdpc_work;
 static void kdpc_worker(unsigned long dummy);
 #else
 static work_struct_t kdpc_work;
-static void kdpc_worker(void *data);
+static void kdpc_worker(worker_param_t dummy);
 #endif
 
 static struct nt_list kdpc_list;
@@ -82,7 +82,7 @@ static struct nt_list bus_driver_list;
 static work_struct_t ntos_work_item_work;
 static struct nt_list ntos_work_item_list;
 static NT_SPIN_LOCK ntos_work_item_list_lock;
-static void ntos_work_item_worker(void *data);
+static void ntos_work_item_worker(worker_param_t dummy);
 
 NT_SPIN_LOCK irp_cancel_lock;
 
@@ -138,11 +138,11 @@ int ntoskernel_init(void)
 
 	thread_event_waitq_pool = NULL;
 	do_gettimeofday(&now);
-	wrap_ticks_to_boot = (u64)now.tv_sec * TICKSPERSEC;
+	wrap_ticks_to_boot = TICKS_1601_TO_1970;
+	wrap_ticks_to_boot += (u64)now.tv_sec * TICKSPERSEC;
 	wrap_ticks_to_boot += now.tv_usec * 10;
-	wrap_ticks_to_boot -= jiffies * TICKSPERSEC / HZ;
-	wrap_ticks_to_boot += TICKS_1601_TO_1970;
-
+	wrap_ticks_to_boot -= jiffies * TICKSPERJIFFY;
+	DBGTRACE2("%Lu", wrap_ticks_to_boot);
 #ifdef USE_OWN_NTOS_WORKQUEUE
 	ntos_wq = create_singlethread_workqueue("ntos_wq");
 #endif
@@ -304,8 +304,7 @@ static void update_user_shared_data_proc(unsigned long data)
 		jiffies * TICKSPERSEC / HZ;
 	*((ULONG64 *)&kuser_shared_data.tick) = jiffies;
 
-	shared_data_timer.expires += 30 * HZ / 1000 + 1;
-	add_timer(&shared_data_timer);
+	mod_timer(&shared_data_timer, jiffies + MSEC_TO_HZ(30));
 }
 #endif
 
@@ -621,6 +620,8 @@ static void timer_proc(unsigned long data)
 	BUG_ON(wrap_timer->wrap_timer_magic != WRAP_TIMER_MAGIC);
 	BUG_ON(nt_timer->wrap_timer_magic != WRAP_TIMER_MAGIC);
 #endif
+	if (wrap_timer->repeat)
+		mod_timer(&wrap_timer->timer, jiffies + wrap_timer->repeat);
 	KeSetEvent((struct nt_event *)nt_timer, 0, FALSE);
 	kdpc = nt_timer->kdpc;
 	if (kdpc && kdpc->func) {
@@ -630,10 +631,6 @@ static void timer_proc(unsigned long data)
 		queue_kdpc(kdpc);
 #endif
 	}
-	nt_spin_lock(&timer_lock);
-	if (wrap_timer->repeat)
-		mod_timer(&wrap_timer->timer, jiffies + wrap_timer->repeat);
-	nt_spin_unlock(&timer_lock);
 	TRACEEXIT5(return);
 }
 
@@ -699,9 +696,7 @@ wstdcall void WIN_FUNC(KeInitializeTimer,1)
 BOOLEAN wrap_set_timer(struct nt_timer *nt_timer, unsigned long expires_hz,
 		       unsigned long repeat_hz, struct kdpc *kdpc)
 {
-	BOOLEAN ret;
 	struct wrap_timer *wrap_timer;
-	KIRQL irql;
 
 	TRACEENTER4("%p, %lu, %lu, %p, %lu",
 		    nt_timer, expires_hz, repeat_hz, kdpc, jiffies);
@@ -724,16 +719,13 @@ BOOLEAN wrap_set_timer(struct nt_timer *nt_timer, unsigned long expires_hz,
 		wrap_timer->wrap_timer_magic = WRAP_TIMER_MAGIC;
 	}
 #endif
-	irql = nt_spin_lock_irql(&timer_lock, DISPATCH_LEVEL);
 	if (kdpc)
 		nt_timer->kdpc = kdpc;
 	wrap_timer->repeat = repeat_hz;
 	if (mod_timer(&wrap_timer->timer, jiffies + expires_hz))
-		ret = TRUE;
+		TRACEEXIT5(return TRUE);
 	else
-		ret = FALSE;
-	nt_spin_unlock_irql(&timer_lock, irql);
-	TRACEEXIT5(return ret);
+		TRACEEXIT5(return FALSE);
 }
 
 wstdcall BOOLEAN WIN_FUNC(KeSetTimerEx,4)
@@ -759,9 +751,7 @@ wstdcall BOOLEAN WIN_FUNC(KeSetTimer,3)
 wstdcall BOOLEAN WIN_FUNC(KeCancelTimer,1)
 	(struct nt_timer *nt_timer)
 {
-	BOOLEAN canceled;
 	struct wrap_timer *wrap_timer;
-	KIRQL irql;
 
 	TRACEENTER5("%p", nt_timer);
 	wrap_timer = nt_timer->wrap_timer;
@@ -770,21 +760,16 @@ wstdcall BOOLEAN WIN_FUNC(KeCancelTimer,1)
 		return TRUE;
 	}
 #ifdef TIMER_DEBUG
-	DBGTRACE5("canceling timer %p", wrap_timer);
 	BUG_ON(wrap_timer->wrap_timer_magic != WRAP_TIMER_MAGIC);
 #endif
-	DBGTRACE5("deleting timer %p(%p)", wrap_timer, nt_timer);
+	DBGTRACE5("canceling timer %p(%p)", wrap_timer, nt_timer);
 	/* disable timer before deleting so if it is periodic timer, it
 	 * won't be re-armed after deleting */
-	irql = nt_spin_lock_irql(&timer_lock, DISPATCH_LEVEL);
 	wrap_timer->repeat = 0;
 	if (del_timer(&wrap_timer->timer))
-		canceled = TRUE;
+		TRACEEXIT5(return TRUE);
 	else
-		canceled = FALSE;
-	nt_spin_unlock_irql(&timer_lock, irql);
-	DBGTRACE5("canceled (%p): %d", wrap_timer, canceled);
-	TRACEEXIT5(return canceled);
+		TRACEEXIT5(return FALSE);
 }
 
 wstdcall BOOLEAN WIN_FUNC(KeReadStateTimer,1)
@@ -804,9 +789,9 @@ wstdcall void WIN_FUNC(KeInitializeDpc,3)
 }
 
 #ifdef KDPC_TASKLET
-static void kdpc_worker(unsigned long data)
+static void kdpc_worker(unsigned long dummy)
 #else
-static void kdpc_worker(void *data)
+static void kdpc_worker(worker_param_t dummy)
 #endif
 {
 	struct nt_list *entry;
@@ -861,6 +846,7 @@ static BOOLEAN queue_kdpc(struct kdpc *kdpc)
 	} else
 		ret = FALSE;
 	nt_spin_unlock_irql(&kdpc_list_lock, irql);
+	DBGTRACE5("%d", ret);
 	TRACEEXIT5(return ret);
 }
 
@@ -878,32 +864,26 @@ static BOOLEAN dequeue_kdpc(struct kdpc *kdpc)
 		ret = TRUE;
 	}
 	nt_spin_unlock_irql(&kdpc_list_lock, irql);
+	DBGTRACE5("%d", ret);
 	return ret;
 }
 
 wstdcall BOOLEAN WIN_FUNC(KeInsertQueueDpc,3)
 	(struct kdpc *kdpc, void *arg1, void *arg2)
 {
-	BOOLEAN ret;
-
 	TRACEENTER5("%p, %p, %p", kdpc, arg1, arg2);
 	kdpc->arg1 = arg1;
 	kdpc->arg2 = arg2;
-	ret = queue_kdpc(kdpc);
-	TRACEEXIT5(return ret);
+	TRACEEXIT5(return queue_kdpc(kdpc));
 }
 
 wstdcall BOOLEAN WIN_FUNC(KeRemoveQueueDpc,1)
 	(struct kdpc *kdpc)
 {
-	BOOLEAN ret;
-
-	TRACEENTER3("%p", kdpc);
-	ret = dequeue_kdpc(kdpc);
-	TRACEEXIT3(return ret);
+	TRACEEXIT3(return dequeue_kdpc(kdpc));
 }
 
-static void ntos_work_item_worker(void *data)
+static void ntos_work_item_worker(worker_param_t dummy)
 {
 	struct ntos_work_item *ntos_work_item;
 	struct nt_list *cur;
@@ -930,7 +910,7 @@ static void ntos_work_item_worker(void *data)
 int schedule_ntos_work_item(NTOS_WORK_FUNC func, void *arg1, void *arg2)
 {
 	struct ntos_work_item *ntos_work_item;
-	KIRQL irql;
+	unsigned long flags;
 
 	WORKENTER("adding work: %p, %p, %p", func, arg1, arg2);
 	ntos_work_item = kmalloc(sizeof(*ntos_work_item), gfp_irql());
@@ -941,9 +921,9 @@ int schedule_ntos_work_item(NTOS_WORK_FUNC func, void *arg1, void *arg2)
 	ntos_work_item->func = func;
 	ntos_work_item->arg1 = arg1;
 	ntos_work_item->arg2 = arg2;
-	irql = nt_spin_lock_irql(&ntos_work_item_list_lock, DISPATCH_LEVEL);
+	nt_spin_lock_irqsave(&ntos_work_item_list_lock, flags);
 	InsertTailList(&ntos_work_item_list, &ntos_work_item->list);
-	nt_spin_unlock_irql(&ntos_work_item_list_lock, irql);
+	nt_spin_unlock_irqrestore(&ntos_work_item_list_lock, flags);
 	schedule_ntos_work(&ntos_work_item_work);
 	WORKEXIT(return 0);
 }
@@ -1017,27 +997,57 @@ wstdcall void *WIN_FUNC(ExAllocatePoolWithTag,3)
 {
 	void *addr;
 
-	TRACEENTER4("pool_type: %d, size: %lu, tag: %u", pool_type,
+	TRACEENTER4("pool_type: %d, size: %lu, tag: 0x%x", pool_type,
 		    size, tag);
-	if (size <= KMALLOC_THRESHOLD)
+	if (size <= (16 * 1024))
 		addr = kmalloc(size, gfp_irql());
 	else {
-		if (current_irql() < DISPATCH_LEVEL)
+		KIRQL irql = current_irql();
+		if (irql > DISPATCH_LEVEL) {
+			WARNING("Windows driver allocating %lu bytes in "
+				"irq context: %ld, %ld, %d", size,
+				in_irq(), in_softirq(), in_atomic());
+			return NULL;
+		}
+		/* Hack alert: Some drivers (at least Atheros)
+		 * allocate large amount of memory during
+		 * Miniport(Query/Set)Information, which runs at
+		 * DISPATCH_LEVEL. This means vmalloc is to be called
+		 * at interrupt context, which is not allowed in
+		 * 2.6.19+ kernels. Ideally, kernel should allow
+		 * __vmalloc in interrupt context, but at least for
+		 * now, we circumvent this problem by lowering IRQL to
+		 * PASSIVE_LEVEL and raising IRQL back to
+		 * DISPATCH_LEVEL. This has been tested to work with
+		 * Atheros PCI driver, but other drivers may break. */
+		if (irql == DISPATCH_LEVEL) {
+			DBGTRACE1("Windows driver allocating %lu bytes in "
+				  "interrupt context: %u, %d, %d, %d, %lu",
+				  size, irql, current_irql(), preempt_count(),
+				  in_atomic(), in_interrupt());
+			lower_irql(PASSIVE_LEVEL);
+			DBGTRACE1("%u, %d, %d, %d, %lu", irql, current_irql(),
+				  preempt_count(), in_atomic(), in_interrupt());
+			if (in_interrupt())
+				addr = NULL;
+			else
+				addr = vmalloc(size);
+			raise_irql(irql);
+			DBGTRACE1("%u, %d, %d, %d, %lu", irql, current_irql(),
+				  preempt_count(), in_atomic(), in_interrupt());
+		} else
 			addr = vmalloc(size);
-		else
-			addr = __vmalloc(size, GFP_ATOMIC | __GFP_HIGHMEM,
-					 PAGE_KERNEL);
 	}
 	DBGTRACE4("addr: %p, %lu", addr, size);
-	TRACEEXIT4(return addr);
+	return addr;
 }
 WIN_FUNC_DECL(ExAllocatePoolWithTag,3)
 
-wstdcall void vfree_nonatomic(void *addr, void *ctx)
+wstdcall void vfree_nonintr(void *addr, void *ctx)
 {
 	vfree(addr);
 }
-WIN_FUNC_DECL(vfree_nonatomic,2)
+WIN_FUNC_DECL(vfree_nonintr,2)
 
 wstdcall void WIN_FUNC(ExFreePoolWithTag,2)
 	(void *addr, ULONG tag)
@@ -1048,7 +1058,7 @@ wstdcall void WIN_FUNC(ExFreePoolWithTag,2)
 		kfree(addr);
 	else {
 		if (in_interrupt())
-			schedule_ntos_work_item(WIN_FUNC_PTR(vfree_nonatomic,2),
+			schedule_ntos_work_item(WIN_FUNC_PTR(vfree_nonintr,2),
 						addr, NULL);
 		else
 			vfree(addr);
@@ -1193,7 +1203,7 @@ wstdcall void WIN_FUNC(ExNotifyCallback,3)
 
 	TRACEENTER3("%p", object);
 	irql = nt_spin_lock_irql(&object->lock, DISPATCH_LEVEL);
-	nt_list_for_each_entry(callback, &object->callback_funcs, list){
+	nt_list_for_each_entry(callback, &object->callback_funcs, list) {
 		LIN2WIN3(callback->func, callback->context, arg1, arg2);
 	}
 	nt_spin_unlock_irql(&object->lock, irql);
@@ -1201,19 +1211,19 @@ wstdcall void WIN_FUNC(ExNotifyCallback,3)
 }
 
 /* check and set signaled state; should be called with dispatcher_lock held */
-/* @grab indicates if the event should be put in not-signaled state
+/* @grab indicates if the event should be grabbed or checked
  * - note that a semaphore may stay in signaled state for multiple
  * 'grabs' if the count is > 1 */
-static int check_grab_signaled_state(struct dispatcher_header *dh,
-				     struct task_struct *thread, int grab)
+static int grab_signaled_state(struct dispatcher_header *dh,
+			       struct task_struct *thread, int grab)
 {
 	EVENTTRACE("%p, %p, %d, %d", dh, thread, grab, dh->signal_state);
 	if (is_mutex_dh(dh)) {
 		struct nt_mutex *nt_mutex;
-		/* either no thread owns the mutex or this thread owns
-		 * it */
 		nt_mutex = container_of(dh, struct nt_mutex, dh);
 		EVENTTRACE("%p, %p", nt_mutex, nt_mutex->owner_thread);
+		/* either no thread owns the mutex or this thread owns
+		 * it */
 		assert(dh->signal_state <= 1);
 		assert(nt_mutex->owner_thread == NULL &&
 		       dh->signal_state == 1);
@@ -1249,8 +1259,7 @@ static void wakeup_threads(struct dispatcher_header *dh)
 			   dh, wb, wb->thread);
 		assert(wb->thread != NULL);
 		assert(wb->object == NULL);
-		if (wb->thread &&
-		    check_grab_signaled_state(dh, wb->thread, 1)) {
+		if (wb->thread && grab_signaled_state(dh, wb->thread, 1)) {
 			struct thread_event_waitq *thread_waitq =
 				wb->thread_waitq;
 			EVENTTRACE("%p: waking up task %p for %p", thread_waitq,
@@ -1325,10 +1334,10 @@ wstdcall NTSTATUS WIN_FUNC(KeWaitForMultipleObjects,8)
 	KIRQL irql;
 
 	thread = current;
-	EVENTENTER("thread: %p count = %d, type: %d, reason = %u, "
-		   "waitmode = %u, alertable = %u, timeout = %p", thread,
-		   count, wait_type, wait_reason, wait_mode, alertable,
-		   timeout);
+	EVENTENTER("thread: %p count: %d, type: %d, reason: %u, "
+		   "waitmode: %u, alertable: %u, timeout: %p, irql: %d",
+		   thread, count, wait_type, wait_reason, wait_mode, alertable,
+		   timeout, current_irql());
 
 	if (count > MAX_WAIT_OBJECTS)
 		EVENTEXIT(return STATUS_INVALID_PARAMETER);
@@ -1356,7 +1365,7 @@ wstdcall NTSTATUS WIN_FUNC(KeWaitForMultipleObjects,8)
 		EVENTTRACE("%p: event %p state: %d",
 			   thread, dh, dh->signal_state);
 		/* wait_type == 1 for WaitAny, 0 for WaitAll */
-		if (check_grab_signaled_state(dh, thread, wait_type)) {
+		if (grab_signaled_state(dh, thread, wait_type)) {
 			if (wait_type == WaitAny) {
 				nt_spin_unlock_irql(&dispatcher_lock, irql);
 				if (count > 1)
@@ -1392,7 +1401,7 @@ wstdcall NTSTATUS WIN_FUNC(KeWaitForMultipleObjects,8)
 			   thread, dh, dh->signal_state);
 		wb[i].object = NULL;
 		wb[i].thread_waitq = thread_waitq;
-		if (check_grab_signaled_state(dh, thread, 1)) {
+		if (grab_signaled_state(dh, thread, 1)) {
 			EVENTTRACE("%p: event %p already signaled: %d",
 				   thread, dh, dh->signal_state);
 			/* mark that we are not waiting on this object */
@@ -1693,7 +1702,7 @@ wstdcall KPRIORITY WIN_FUNC(KeQueryPriorityThread,1)
 wstdcall ULONGLONG WIN_FUNC(KeQueryInterruptTime,0)
 	(void)
 {
-	TRACEEXIT5(return jiffies * TICKSPERSEC / HZ);
+	TRACEEXIT5(return jiffies * TICKSPERJIFFY);
 }
 
 wstdcall ULONG WIN_FUNC(KeQueryTimeIncrement,0)
@@ -1706,7 +1715,7 @@ wstdcall void WIN_FUNC(KeQuerySystemTime,1)
 	(LARGE_INTEGER *time)
 {
 	*time = ticks_1601();
-	return;
+	DBGTRACE5("%Lu, %lu", *time, jiffies);
 }
 
 wstdcall void WIN_FUNC(KeQueryTickCount,1)
@@ -1759,38 +1768,7 @@ wstdcall KPRIORITY WIN_FUNC(KeSetPriorityThread,2)
 	return old_prio;
 }
 
-struct trampoline_context {
-	void (*start_routine)(void *) wstdcall;
-	void *context;
-	struct nt_thread *thread;
-};
-
-static int thread_trampoline(void *data)
-{
-	struct trampoline_context ctx;
-	struct nt_thread *thread;
-
-	memcpy(&ctx, data, sizeof(ctx));
-	kfree(data);
-	thread = ctx.thread;
-	thread->task = current;
-	thread->pid = thread->task->pid;
-#ifdef PF_NOFREEZE
-	current->flags |= PF_NOFREEZE;
-#endif
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
-	strncpy(current->comm, "windisdrvr", sizeof(current->comm));
-	current->comm[sizeof(current->comm)-1] = 0;
-#endif
-	DBGTRACE2("thread: %p, task: %p (%d)", thread, thread->task,
-		  thread->pid);
-	LIN2WIN1(ctx.start_routine, ctx.context);
-	DBGTRACE2("thread: %p, task: %p (%d)", thread, thread->task,
-		  thread->pid);
-	return 0;
-}
-
-static struct nt_thread *create_nt_thread(struct task_struct *task)
+static struct nt_thread *create_nt_thread(void)
 {
 	struct nt_thread *thread;
 
@@ -1799,17 +1777,13 @@ static struct nt_thread *create_nt_thread(struct task_struct *task)
 		ERROR("couldn't allocate thread object");
 		return NULL;
 	}
-	thread->task = task;
-	if (task)
-		thread->pid = task->pid;
-	else
-		thread->pid = 0;
+	thread->task = NULL;
+	thread->pid = 0;
 	nt_spin_lock_init(&thread->lock);
 	InitializeListHead(&thread->irps);
 	initialize_dh(&thread->dh, ThreadObject, 0);
 	thread->dh.size = sizeof(*thread);
-	DBGTRACE2("thread: %p, task: %p, pid: %d",
-		  thread, thread->task, thread->pid);
+	DBGTRACE2("thread: %p", thread);
 	return thread;
 }
 
@@ -1841,83 +1815,101 @@ static void remove_nt_thread(struct nt_thread *thread)
 struct nt_thread *get_current_nt_thread(void)
 {
 	struct task_struct *task = current;
-	struct nt_thread *ret;
+	struct nt_thread *thread;
 	struct common_object_header *header;
 	KIRQL irql;
 
-	DBGTRACE5("task: %p", task);
-	ret = NULL;
+	DBGTRACE6("task: %p", task);
+	thread = NULL;
 	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
 	nt_list_for_each_entry(header, &object_list, list) {
-		struct nt_thread *thread;
-		DBGTRACE5("header: %p, type: %d", header, header->type);
+		DBGTRACE6("header: %p, type: %d", header, header->type);
 		if (header->type != OBJECT_TYPE_NT_THREAD)
 			break;
 		thread = HEADER_TO_OBJECT(header);
-		DBGTRACE5("thread: %p, task: %p", thread, thread->task);
-		if (thread->task == task) {
-			ret = thread;
+		DBGTRACE6("thread: %p, task: %p", thread, thread->task);
+		if (thread->task == task)
 			break;
-		}
+		else
+			thread = NULL;
 	}
 	nt_spin_unlock_irql(&ntoskernel_lock, irql);
-	if (ret == NULL)
-		DBGTRACE3("couldn't find thread for task %p, %d",
-			  task, current->pid);
-	DBGTRACE5("current thread = %p", ret);
-	return ret;
+	if (thread == NULL)
+		DBGTRACE6("couldn't find thread for task %p, %d",
+			  task, task->pid);
+	DBGTRACE6("thread: %p", thread);
+	return thread;
+}
+
+struct thread_trampoline_info {
+	void (*func)(void *) wstdcall;
+	void *ctx;
+	struct nt_thread *thread;
+	struct completion started;
+};
+
+static int thread_trampoline(void *data)
+{
+	struct thread_trampoline_info *thread_info = data;
+	typeof(thread_info->func) func;
+	void *ctx;
+
+	func = thread_info->func;
+	ctx = thread_info->ctx;
+	thread_info->thread->task = current;
+	thread_info->thread->pid = current->pid;
+	DBGTRACE2("thread: %p, task: %p (%d)", thread_info->thread,
+		  current, current->pid);
+	complete(&thread_info->started);
+
+#ifdef PF_NOFREEZE
+	current->flags |= PF_NOFREEZE;
+#endif
+	strncpy(current->comm, "windisdrvr", sizeof(current->comm));
+	current->comm[sizeof(current->comm)-1] = 0;
+	LIN2WIN1(func, ctx);
+	ERROR("task: %p", current);
+	return 0;
 }
 
 wstdcall NTSTATUS WIN_FUNC(PsCreateSystemThread,7)
 	(void **phandle, ULONG access, void *obj_attr, void *process,
-	 void *client_id, void (*start_routine)(void *) wstdcall, void *context)
+	 void *client_id, void (*func)(void *) wstdcall, void *ctx)
 {
-	struct trampoline_context *ctx;
-	struct nt_thread *thread;
-	struct task_struct *task;
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
-	int pid;
-#endif
+	struct thread_trampoline_info thread_info;
+	no_warn_unused struct task_struct *task;
+	no_warn_unused int pid;
 
 	TRACEENTER2("phandle = %p, access = %u, obj_attr = %p, process = %p, "
-	            "client_id = %p, start_routine = %p, context = %p",
-	            phandle, access, obj_attr, process, client_id,
-	            start_routine, context);
+	            "client_id = %p, func = %p, context = %p", phandle, access,
+		    obj_attr, process, client_id, func, ctx);
 
-	ctx = kmalloc(sizeof(struct trampoline_context), GFP_KERNEL);
-	if (!ctx)
+	thread_info.thread = create_nt_thread();
+	if (!thread_info.thread)
 		TRACEEXIT2(return STATUS_RESOURCES);
-	ctx->start_routine = start_routine;
-	ctx->context = context;
-	thread = create_nt_thread(NULL);
-	if (!thread) {
-		kfree(ctx);
-		TRACEEXIT2(return STATUS_RESOURCES);
-	}
-	ctx->thread = thread;
+	thread_info.func = func;
+	thread_info.ctx = ctx;
+	init_completion(&thread_info.started);
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
-	pid = kernel_thread(thread_trampoline, ctx,
-			    CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
+	pid = kernel_thread(thread_trampoline, &thread_info, CLONE_SIGHAND);
 	DBGTRACE2("pid = %d", pid);
 	if (pid < 0) {
-		kfree(ctx);
-		free_object(thread);
+		free_object(thread_info.thread);
 		TRACEEXIT2(return STATUS_FAILURE);
 	}
-	task = NULL;
-	DBGTRACE2("created task: %p (%d)", task, pid);
+	DBGTRACE2("created task: %d", pid);
 #else
-	task = KTHREAD_RUN(thread_trampoline, ctx, "windisdrvr");
+	task = KTHREAD_RUN(thread_trampoline, &thread_info, "windisdrvr");
 	if (IS_ERR(task)) {
-		kfree(ctx);
-		free_object(thread);
+		free_object(thread_info.thread);
 		TRACEEXIT2(return STATUS_FAILURE);
 	}
 	DBGTRACE2("created task: %p (%d)", task, task->pid);
 #endif
-	*phandle = OBJECT_TO_HEADER(thread);
-	DBGTRACE2("created thread: %p, %p", thread, *phandle);
+	wait_for_completion(&thread_info.started);
+	*phandle = OBJECT_TO_HEADER(thread_info.thread);
+	DBGTRACE2("created thread: %p, %p", thread_info.thread, *phandle);
 	TRACEEXIT2(return STATUS_SUCCESS);
 }
 
@@ -1982,7 +1974,7 @@ wstdcall void *WIN_FUNC(MmAllocateContiguousMemorySpecifyCache,5)
 {
 	void *addr;
 	size_t page_length = ((size + PAGE_SIZE - 1) / PAGE_SIZE) * PAGE_SIZE;
-	DBGTRACE2("%lu, %u, %Lu, %Lu, %Lu, %d", size, page_length,
+	DBGTRACE2("%lu, %zu, %Lu, %Lu, %Lu, %d", size, page_length,
 		  lowest, highest, boundary, cache_type);
 	addr = ExAllocatePoolWithTag(NonPagedPool, page_length, 0);
 	DBGTRACE2("%p", addr);
@@ -2377,7 +2369,7 @@ wstdcall NTSTATUS WIN_FUNC(ZwReadFile,9)
 	}
 	fo = HANDLE_TO_OBJECT(coh);
 	file = fo->wrap_bin_file;
-	DBGTRACE2("file: %s (%u)", file->name, file->size);
+	DBGTRACE2("file: %s (%zu)", file->name, file->size);
 	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
 	if (byte_offset)
 		offset = *byte_offset;
@@ -2412,7 +2404,7 @@ wstdcall NTSTATUS WIN_FUNC(ZwWriteFile,9)
 	}
 	fo = HANDLE_TO_OBJECT(coh);
 	file = fo->wrap_bin_file;
-	DBGTRACE2("file: %u, %u", file->size, length);
+	DBGTRACE2("file: %zu, %u", file->size, length);
 	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
 	if (byte_offset)
 		offset = *byte_offset;
@@ -2566,6 +2558,13 @@ wstdcall NTSTATUS WIN_FUNC(ZwQueryValueKey,6)
 	return STATUS_INVALID_PARAMETER;
 }
 
+wstdcall NTSTATUS WIN_FUNC(ZwDeleteKey,1)
+	(void *handle)
+{
+	TRACEENTER2("%p", handle);
+	return STATUS_SUCCESS;
+}
+
 wstdcall NTSTATUS WIN_FUNC(WmiSystemControl,4)
 	(struct wmilib_context *info, struct device_object *dev_obj,
 	 struct irp *irp, void *irp_disposition)
