@@ -43,6 +43,7 @@
 #include <linux/ethtool.h>
 #include <linux/if_arp.h>
 #include <linux/rtnetlink.h>
+#include <linux/highmem.h>
 
 #include "winnt_types.h"
 
@@ -203,29 +204,40 @@ typedef struct {
 	workqueue_struct_t *workq;
 } work_struct_t;
 
-#define initialize_work(work_struct, worker_func, worker_data)	\
+#define initialize_work(work, pfunc, pdata)			\
 	do {							\
-		(work_struct)->func = worker_func;		\
-		(work_struct)->data = worker_data;		\
-		(work_struct)->workq = NULL;			\
+		(work)->func = (pfunc);				\
+		(work)->data = (pdata);				\
+		(work)->workq = NULL;				\
 	} while (0)
 
 #undef create_singlethread_workqueue
 #define create_singlethread_workqueue wrap_create_wq
 #define destroy_workqueue wrap_destroy_wq
 #define queue_work wrap_queue_work
-#define cancel_delayed_work wrap_cancel_delayed_work
 
 workqueue_struct_t *wrap_create_wq(const char *name);
 void wrap_destroy_wq(workqueue_struct_t *workq);
 void wrap_queue_work(workqueue_struct_t *workq, work_struct_t *work) wfastcall;
-void wrap_cancel_delayed_work(work_struct_t *work);
+void wrap_cancel_work(work_struct_t *work);
+typedef void *worker_param_t;
+#define worker_param_data(param, type, member) param
 
 #else // USE_OWN_WQ
 
 typedef struct workqueue_struct workqueue_struct_t;
 typedef struct work_struct work_struct_t;
-#define initialize_work INIT_WORK
+
+#ifdef INIT_WORK_NAR
+#define initialize_work(work, func, data) INIT_WORK(work, func)
+typedef struct work_struct *worker_param_t;
+#define worker_param_data(param, type, member)	\
+	container_of(param, type, member)
+#else
+#define initialize_work(work, func, data) INIT_WORK(work, func, data)
+typedef void *worker_param_t;
+#define worker_param_data(param, type, member) param
+#endif // INIT_WORK_NAR
 
 #endif // USE_OWN_WQ
 
@@ -354,6 +366,10 @@ typedef u32 pm_message_t;
 #define ISR_PT_REGS_ARG , NULL
 #endif
 
+#ifndef flush_icache_range
+#define flush_icache_range(start, end) do { } while (0)
+#endif
+
 #define memcpy_skb(skb, from, length)			\
 	memcpy(skb_put(skb, length), from, length)
 
@@ -386,12 +402,11 @@ typedef u32 pm_message_t;
 #undef CONFIG_USB_MODULE
 #endif
 
-#define KMALLOC_THRESHOLD 130000
-
 /* TICK is 100ns */
 #define TICKSPERSEC		10000000LL
 #define TICKSPERMSEC		10000
 #define SECSPERDAY		86400
+#define TICKSPERJIFFY		((10000000 + HZ - 1) / HZ)
 
 /* 1601 to 1970 is 369 years plus 89 leap days */
 #define SECS_1601_TO_1970	((369 * 365 + 89) * (u64)SECSPERDAY)
@@ -410,7 +425,7 @@ extern u64 wrap_ticks_to_boot;
 
 static inline u64 ticks_1601(void)
 {
-	return wrap_ticks_to_boot + (u64)jiffies * TICKSPERSEC / HZ;
+	return wrap_ticks_to_boot + (u64)jiffies * TICKSPERJIFFY;
 }
 
 typedef void (*generic_func)(void);
@@ -504,8 +519,6 @@ struct wrap_driver {
 	struct wrap_ndis_driver *ndis_driver;
 };
 
-struct usbd_pipe_information;
-
 struct wrap_device {
 	/* first part is (de)initialized once by loader */
 	struct nt_list list;
@@ -713,8 +726,8 @@ NTSTATUS IoPassIrpDown(struct device_object *dev_obj, struct irp *irp) wstdcall;
 WIN_FUNC_DECL(IoPassIrpDown,2)
 NTSTATUS IoSyncForwardIrp(struct device_object *dev_obj,
 			  struct irp *irp) wstdcall;
-NTSTATUS IoAsyncForwardIrp (struct device_object *dev_obj,
-			    struct irp *irp) wstdcall;
+NTSTATUS IoAsyncForwardIrp(struct device_object *dev_obj,
+			   struct irp *irp) wstdcall;
 NTSTATUS IoInvalidDeviceRequest(struct device_object *dev_obj,
 				struct irp *irp) wstdcall;
 
@@ -767,11 +780,11 @@ void adjust_user_shared_data_addr(char *driver, unsigned long length);
 static inline KIRQL current_irql(void)
 {
 	if (in_irq() || irqs_disabled())
-		TRACEEXIT5(return DEVICE_LEVEL);
-	else if (in_atomic())
-		TRACEEXIT5(return DISPATCH_LEVEL);
+		TRACEEXIT6(return DEVICE_LEVEL);
+	else if (in_interrupt())
+		TRACEEXIT6(return DISPATCH_LEVEL);
 	else
-		TRACEEXIT5(return PASSIVE_LEVEL);
+		TRACEEXIT6(return PASSIVE_LEVEL);
 }
 
 static inline KIRQL raise_irql(KIRQL newirql)
@@ -782,14 +795,14 @@ static inline KIRQL raise_irql(KIRQL newirql)
 		local_bh_disable();
 		preempt_disable();
 	}
-	DBGTRACE5("%d, %d", irql, newirql);
+	DBGTRACE6("%d, %d", irql, newirql);
 	return irql;
 }
 
 static inline void lower_irql(KIRQL oldirql)
 {
 	KIRQL irql = current_irql();
-	DBGTRACE5("%d, %d", irql, oldirql);
+	DBGTRACE6("%d, %d", irql, oldirql);
 	DBG_BLOCK(2) {
 		if (irql < oldirql)
 			ERROR("invalid irql: %d < %d", irql, oldirql);
@@ -889,28 +902,28 @@ do {									\
 } while (0)
 
 #define atomic_unary_op(var, size, oper)			\
-	do {							\
-		if (size == 1)					\
-			__asm__ __volatile__(			\
-				LOCK_PREFIX oper "b %b0\n\t"	\
-				: "+m" (var));			\
-		else if (size == 2)				\
-			__asm__ __volatile__(			\
-				LOCK_PREFIX oper "w %w0\n\t"	\
-				: "+m" (var));			\
-		else if (size == 4)				\
-			__asm__ __volatile__(			\
-				LOCK_PREFIX oper "l %0\n\t"	\
-				: "+m" (var));			\
-		else if (size == 8)				\
-			__asm__ __volatile__(			\
-				LOCK_PREFIX oper "q %q0\n\t"	\
-				: "+m" (var));			\
-		else {						\
-			extern void _invalid_op_size_(void);	\
-			_invalid_op_size_();			\
-		}						\
-	} while (0)
+do {								\
+	if (size == 1)						\
+		__asm__ __volatile__(				\
+			LOCK_PREFIX oper "b %b0\n\t"		\
+			: "+m" (var));				\
+	else if (size == 2)					\
+		__asm__ __volatile__(				\
+			LOCK_PREFIX oper "w %w0\n\t"		\
+			: "+m" (var));				\
+	else if (size == 4)					\
+		__asm__ __volatile__(				\
+			LOCK_PREFIX oper "l %0\n\t"		\
+			: "+m" (var));				\
+	else if (size == 8)					\
+		__asm__ __volatile__(				\
+			LOCK_PREFIX oper "q %q0\n\t"		\
+			: "+m" (var));				\
+	else {							\
+		extern void _invalid_op_size_(void);		\
+		_invalid_op_size_();				\
+	}							\
+} while (0)
 
 #define atomic_inc_var_size(var, size) atomic_unary_op(var, size, "inc")
 
@@ -932,7 +945,7 @@ do {									\
 
 #define post_atomic_add(var, i) (pre_atomic_add(var, i) + i)
 
-#define atomic_insert_list_head(head, oldhead, newhead)			\
+#define atomic_insert_list_head(oldhead, head, newhead)			\
 	do {								\
 		oldhead = head;						\
 	} while (cmpxchg(&(head), oldhead, newhead) != oldhead)
