@@ -31,7 +31,7 @@ static unsigned int urb_id = 0;
 #define DUMP_URB_BUFFER(urb, dir)					\
 	while (debug >= 2) {						\
 		int i;							\
-		char msg[100], *t;					\
+		char msg[20], *t;					\
 		if (!urb->transfer_buffer)				\
 			break;						\
 		if (!((usb_pipein(urb->pipe) && dir == USB_DIR_IN) ||	\
@@ -137,7 +137,7 @@ static struct tasklet_struct wrap_urb_complete_work;
 static void wrap_urb_complete_worker(unsigned long dummy);
 #else
 static work_struct_t wrap_urb_complete_work;
-static void wrap_urb_complete_worker(void *dummy);
+static void wrap_urb_complete_worker(worker_param_t dummy);
 #endif
 
 static void wrap_free_urb(struct urb *urb);
@@ -161,8 +161,6 @@ void usb_exit(void)
 {
 #ifdef USB_TASKLET
 	tasklet_kill(&wrap_urb_complete_work);
-#else
-	cancel_delayed_work(&wrap_urb_complete_work);
 #endif
 	USBEXIT(return);
 }
@@ -248,7 +246,7 @@ static NTSTATUS nt_urb_irp_status(USBD_STATUS nt_urb_status)
 	case USBD_STATUS_SUCCESS:
 		return STATUS_SUCCESS;
 	case USBD_STATUS_DEVICE_GONE:
-		return STATUS_DEVICE_REMOVED;
+		return STATUS_DEVICE_NOT_CONNECTED;
 	case USBD_STATUS_PENDING:
 		return STATUS_PENDING;
 	case USBD_STATUS_NOT_SUPPORTED:
@@ -520,7 +518,7 @@ static void wrap_urb_complete(struct urb *urb ISR_PT_REGS_PARAM_DECL)
 #ifdef USB_TASKLET
 static void wrap_urb_complete_worker(unsigned long dummy)
 #else
-static void wrap_urb_complete_worker(void *dummy)
+static void wrap_urb_complete_worker(worker_param_t dummy)
 #endif
 {
 	struct irp *irp;
@@ -618,7 +616,7 @@ static USBD_STATUS wrap_bulk_or_intr_trans(struct irp *irp)
 	udev = irp->wd->usb.udev;
 	bulk_int_tx = &nt_urb->bulk_int_transfer;
 	pipe_handle = bulk_int_tx->pipe_handle;
-	USBTRACE("flags = %X, length = %u, buffer = %p, handle: %p",
+	USBTRACE("flags: %X, length: %u, buffer: %p, handle: %p",
 		 bulk_int_tx->transfer_flags,
 		 bulk_int_tx->transfer_buffer_length,
 		 bulk_int_tx->transfer_buffer, pipe_handle);
@@ -685,11 +683,10 @@ static USBD_STATUS wrap_vendor_or_class_req(struct irp *irp)
 	nt_urb = URB_FROM_IRP(irp);
 	udev = irp->wd->usb.udev;
 	vc_req = &nt_urb->vendor_class_request;
-	USBTRACE("bits = %x, req = %x, val = %08x, index = %08x, flags = %x,"
-		 "tx_buf = %p, tx_buf_len = %d", vc_req->reserved_bits,
-		 vc_req->request, vc_req->value, vc_req->index,
-		 vc_req->transfer_flags, vc_req->transfer_buffer,
-		 vc_req->transfer_buffer_length);
+	USBTRACE("bits: %x, req: %x, val: %08x, index: %08x, flags: %x,"
+		 "buf: %p, len: %d", vc_req->reserved_bits, vc_req->request,
+		 vc_req->value, vc_req->index, vc_req->transfer_flags,
+		 vc_req->transfer_buffer, vc_req->transfer_buffer_length);
 
 	USBTRACE("%x", nt_urb->header.function);
 	switch (nt_urb->header.function) {
@@ -762,8 +759,7 @@ static USBD_STATUS wrap_vendor_or_class_req(struct irp *irp)
 	dr->wLength = cpu_to_le16((u16)urb->transfer_buffer_length);
 
 	usb_fill_control_urb(urb, udev, pipe, (unsigned char *)dr,
-			     urb->transfer_buffer,
-			     urb->transfer_buffer_length,
+			     urb->transfer_buffer, urb->transfer_buffer_length,
 			     wrap_urb_complete, urb->context);
 	status = wrap_submit_urb(irp);
 	USBTRACE("status: %08X", status);
@@ -831,6 +827,7 @@ static USBD_STATUS wrap_abort_pipe(struct usb_device *udev, struct irp *irp)
 		}
 	}
 	IoReleaseCancelSpinLock(irql);
+	NT_URB_STATUS(nt_urb) = USBD_STATUS_CANCELED;
 	USBEXIT(return USBD_STATUS_SUCCESS);
 }
 
@@ -1087,7 +1084,6 @@ static USBD_STATUS wrap_process_nt_urb(struct irp *irp)
 
 	case URB_FUNCTION_ABORT_PIPE:
 		status = wrap_abort_pipe(udev, irp);
-		NT_URB_STATUS(nt_urb) = status;
 		break;
 
 	default:
@@ -1341,6 +1337,21 @@ WIN_FUNC(USBD_ParseConfigurationDescriptor,3)
 						   -1, -1, -1);
 }
 
+wstdcall usb_common_descriptor_t *WIN_FUNC(USBD_ParseDescriptors,4)
+	(void *buf, ULONG length, void *start, LONG type)
+{
+	usb_common_descriptor_t *descr = start;
+
+	while ((void *)descr < buf + length) {
+		if (descr->bDescriptorType == type)
+			return descr;
+		if (descr->bLength == 0)
+			break;
+		descr = (void *)descr + descr->bLength;
+	}
+	USBEXIT(return NULL);
+}
+
 wstdcall void WIN_FUNC(USBD_GetUSBDIVersion,1)
 	(struct usbd_version_info *version_info)
 {
