@@ -41,19 +41,16 @@ static int workq_thread(void *data)
 #ifdef PF_NOFREEZE
 	current->flags |= PF_NOFREEZE;
 	set_user_nice(current, -5);
+#else
+	sigfillset(&current->blocked);
+	current->nice -= 5;
 #endif
 	while (1) {
 		if (wait_event_interruptible(workq->waitq_head,
 					     workq->pending)) {
 			/* TODO: deal with signal */
-			spin_lock_irq(SIG_LOCK(current));
+			WARNING("signal not blocked?");
 			flush_signals(current);
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
-			recalc_sigpending(current);
-#else
-			recalc_sigpending();
-#endif
-			spin_unlock_irq(SIG_LOCK(current));
 			continue;
 		}
 		spin_lock_irqsave(&workq->lock, flags);
@@ -64,19 +61,18 @@ static int workq_thread(void *data)
 		else {
 			struct list_head *entry = workq->work_list.next;
 			work = list_entry(entry, work_struct_t, list);
-			list_del(entry);
-			if (work->workq) {
-				assert(work->workq == workq);
-				work->workq = NULL;
-			} else
+			BUG_ON(work->workq != workq);
+			if (xchg(&work->workq, NULL))
+				list_del(entry);
+			else
 				work = NULL;
 		}
 		spin_unlock_irqrestore(&workq->lock, flags);
 		if (work)
 			work->func(work->data);
 	}
-	/* set workq for each work to NULL so if cancel_delayed_work
-	 * is called later, it won't access workq */
+	/* set workq for each work to NULL so if work is cancelled
+	 * later, it won't access workq */
 	list_for_each_entry(work, &workq->work_list, list) {
 		work->workq = NULL;
 	}
@@ -100,7 +96,7 @@ wfastcall void wrap_queue_work(workqueue_struct_t *workq, work_struct_t *work)
 	spin_unlock_irqrestore(&workq->lock, flags);
 }
 
-void wrap_cancel_delayed_work(work_struct_t *work)
+void wrap_cancel_work(work_struct_t *work)
 {
 	workqueue_struct_t *workq;
 	unsigned long flags;
