@@ -18,6 +18,7 @@
 #include "pnp.h"
 #include "loader.h"
 #include "wrapndis.h"
+#include <linux/inetdevice.h>
 
 extern char *if_name;
 extern int hangcheck_interval;
@@ -274,15 +275,14 @@ static NDIS_STATUS miniport_init(struct wrap_ndis_device *wnd)
 	 * fail after boot */
 	sleep_hz(HZ / 2);
 	set_bit(HW_INITIALIZED, &wnd->hw_status);
-	hangcheck_add(wnd);
 	/* the description about NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND is
 	 * misleading/confusing; we just ignore it */
 	status = miniport_query_info(wnd, OID_PNP_CAPABILITIES,
 				     &pnp_capa, sizeof(pnp_capa));
 	if (status == NDIS_STATUS_SUCCESS)
-		wnd->pm_capa = TRUE;
+		wnd->attributes |= NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND;
 	else
-		wnd->pm_capa = FALSE;
+		wnd->attributes &= ~NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND;
 	DBGTRACE1("%d", pnp_capa.wakeup_capa.min_magic_packet_wakeup);
 	TRACEEXIT1(return NDIS_STATUS_SUCCESS);
 }
@@ -373,7 +373,7 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 		hangcheck_del(wnd);
 		del_stats_timer(wnd);
 		status = NDIS_STATUS_NOT_SUPPORTED;
-		if (wnd->pm_capa == TRUE) {
+		if (wnd->attributes & NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND) {
 			enum ndis_power_state pm_state = state;
 			if (wnd->ndis_wolopts) {
 				status = miniport_set_int(wnd,
@@ -420,9 +420,9 @@ static int ndis_set_mac_address(struct net_device *dev, void *p)
 
 	memcpy(mac, addr->sa_data, sizeof(mac));
 	memset(mac_string, 0, sizeof(mac_string));
-	res = snprintf(mac_string, sizeof(mac_string), MACSTR, MAC2STR(mac));
+	res = snprintf(mac_string, sizeof(mac_string), MACSTRSEP, MAC2STR(mac));
 	DBGTRACE1("%d", res);
-	if (res != (2 * sizeof(mac)))
+	if (res != (2 * sizeof(mac) + sizeof(mac) - 1))
 		TRACEEXIT1(return -EINVAL);
 
 	RtlInitAnsiString(&ansi, mac_string);
@@ -431,7 +431,7 @@ static int ndis_set_mac_address(struct net_device *dev, void *p)
 		TRACEEXIT1(return -EINVAL);
 	}
 	param.type = NdisParameterString;
-	RtlInitAnsiString(&ansi, "NetworkAddress");
+	RtlInitAnsiString(&ansi, "mac_address");
 	if (RtlAnsiStringToUnicodeString(&key, &ansi, TRUE))
 		TRACEEXIT1(return -EINVAL);
 	NdisWriteConfiguration(&res, wnd->nmb, &key, &param);
@@ -463,6 +463,7 @@ static struct ndis_packet *alloc_tx_packet(struct wrap_ndis_device *wnd,
 	NdisAllocatePacket(&status, &packet, wnd->tx_packet_pool);
 	if (status != NDIS_STATUS_SUCCESS)
 		return NULL;
+	packet->private.flags |= NDIS_MAC_OPTION_NO_LOOPBACK;
 	NdisAllocateBuffer(&status, &buffer, wnd->tx_buffer_pool,
 			   skb->data, skb->len);
 	if (status != NDIS_STATUS_SUCCESS) {
@@ -623,11 +624,12 @@ static int miniport_tx_packets(struct wrap_ndis_device *wnd, int start, int n)
 	TRACEEXIT3(return sent);
 }
 
-static void tx_worker(void *param)
+static void tx_worker(worker_param_t param)
 {
-	struct wrap_ndis_device *wnd = param;
+	struct wrap_ndis_device *wnd;
 	int n;
 
+	wnd = worker_param_data(param, struct wrap_ndis_device, tx_work);
 	TRACEENTER3("tx_ok %d", wnd->tx_ok);
 	while (wnd->tx_ok) {
 		if (down_interruptible(&wnd->tx_ring_mutex))
@@ -757,6 +759,26 @@ static int ndis_net_dev_close(struct net_device *net_dev)
 	return 0;
 }
 
+static int ndis_change_mtu(struct net_device *net_dev, int mtu)
+{
+	struct wrap_ndis_device *wnd = netdev_priv(net_dev);
+	int max;
+
+	if (mtu < ETH_ZLEN)
+		return -EINVAL;
+	if (miniport_query_int(wnd, OID_GEN_MAXIMUM_TOTAL_SIZE, &max) !=
+	    NDIS_STATUS_SUCCESS)
+		return -EOPNOTSUPP;
+	DBGTRACE2("%d", max);
+	max -= ETH_HLEN;
+	if (max <= ETH_ZLEN)
+		return -EINVAL;
+	if (mtu + ETH_HLEN > max)
+		return -EINVAL;
+	net_dev->mtu = mtu;
+	return 0;
+}
+
 #ifdef CONFIG_NET_POLL_CONTROLLER
 static void ndis_poll_controller(struct net_device *dev)
 {
@@ -833,7 +855,7 @@ static int ndis_set_wol(struct net_device *dev, struct ethtool_wolinfo *wol)
 
 	if (!(wol->wolopts & WAKE_MAGIC))
 		return -EINVAL;
-	if (!wnd->pm_capa)
+	if (!(wnd->attributes & NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND))
 		return -EOPNOTSUPP;
 	status = miniport_query_info(wnd, OID_PNP_CAPABILITIES,
 				     &pnp_capa, sizeof(pnp_capa));
@@ -861,8 +883,12 @@ static int notifier_event(struct notifier_block *notifier, unsigned long event,
 			  void *ptr)
 {
 	struct net_device *net_dev = (struct net_device *)ptr;
-	struct wrap_ndis_device *wnd = netdev_priv(net_dev);
+	struct wrap_ndis_device *wnd;
 
+	if (net_dev->open != ndis_net_dev_open)
+		return NOTIFY_DONE;
+
+	wnd = netdev_priv(net_dev);
 	/* called with rtnl lock held, so no need to lock */
 	switch (event) {
 	case NETDEV_CHANGENAME:
@@ -873,6 +899,9 @@ static int notifier_event(struct notifier_block *notifier, unsigned long event,
 		       sizeof(wnd->netdev_name));
 		wrap_procfs_add_ndis_device(wnd);
 		break;
+	default:
+		DBGTRACE2("%lx", event);
+		break;
 	}
 	return NOTIFY_DONE;
 }
@@ -978,10 +1007,10 @@ static void link_status_handler(struct wrap_ndis_device *wnd)
 	union iwreq_data wrqu;
 	NDIS_STATUS res;
 	const int assoc_size = sizeof(*ndis_assoc_info) + IW_CUSTOM_MAX + 32;
+	int i;
 #if WIRELESS_EXT <= 17
 	unsigned char *wpa_assoc_info, *ies;
 	unsigned char *p;
-	int i;
 #endif
 
 	TRACEENTER2("link: %d", netif_carrier_ok(wnd->net_dev));
@@ -994,6 +1023,20 @@ static void link_status_handler(struct wrap_ndis_device *wnd)
 		memset(&wrqu, 0, sizeof(wrqu));
 		wrqu.ap_addr.sa_family = ARPHRD_ETHER;
 		wireless_send_event(wnd->net_dev, SIOCGIWAP, &wrqu, NULL);
+		/* In IBSS (ad-hoc) mode, it may be desirable to have
+		 * one node configured for association, but drivers
+		 * disassociate if last node disassociates; to
+		 * configure again, set essid */
+		if (wnd->infrastructure_mode == Ndis802_11IBSS &&
+		    wnd->essid.length > 0) {
+			set_essid(wnd, wnd->essid.essid, wnd->essid.length);
+			for (i = 0; i < MAX_ENCR_KEYS; i++) {
+				if (wnd->encr_info.keys[i].length <= 0)
+					continue;
+				add_wep_key(wnd, wnd->encr_info.keys[i].key,
+					    wnd->encr_info.keys[i].length, i);
+			}
+		}
 		TRACEEXIT2(return);
 	}
 
@@ -1076,8 +1119,7 @@ static void stats_timer_proc(unsigned long data)
 		TRACEEXIT2(return);
 	set_bit(COLLECT_STATS, &wnd->wrap_ndis_pending_work);
 	schedule_wrap_work(&wnd->wrap_ndis_work);
-	wnd->stats_timer.expires += wnd->stats_interval;
-	add_timer(&wnd->stats_timer);
+	mod_timer(&wnd->stats_timer, jiffies + wnd->stats_interval);
 }
 
 static void add_stats_timer(struct wrap_ndis_device *wnd)
@@ -1088,8 +1130,7 @@ static void add_stats_timer(struct wrap_ndis_device *wnd)
 		wnd->stats_interval *= -1;
 	wnd->stats_timer.data = (unsigned long)wnd;
 	wnd->stats_timer.function = stats_timer_proc;
-	wnd->stats_timer.expires = jiffies + wnd->stats_interval;
-	add_timer(&wnd->stats_timer);
+	mod_timer(&wnd->stats_timer, jiffies + wnd->stats_interval);
 }
 
 static void del_stats_timer(struct wrap_ndis_device *wnd)
@@ -1117,8 +1158,7 @@ static void hangcheck_proc(unsigned long data)
 		set_bit(MINIPORT_RESET, &wnd->wrap_ndis_pending_work);
 		schedule_wrap_work(&wnd->wrap_ndis_work);
 	}
-	wnd->hangcheck_timer.expires += wnd->hangcheck_interval;
-	add_timer(&wnd->hangcheck_timer);
+	mod_timer(&wnd->hangcheck_timer, jiffies + wnd->hangcheck_interval);
 	TRACEEXIT3(return);
 }
 
@@ -1126,16 +1166,19 @@ void hangcheck_add(struct wrap_ndis_device *wnd)
 {
 	if (!wnd->wd->driver->ndis_driver->miniport.hangcheck ||
 	    hangcheck_interval < 0)
+		TRACEEXIT2(return);
+	if (timer_pending(&wnd->hangcheck_timer)) {
+		WARNING("timer is already running!");
 		return;
+	}
 	if (hangcheck_interval > 0)
 		wnd->hangcheck_interval = hangcheck_interval * HZ;
 	if (wnd->hangcheck_interval < 0)
 		wnd->hangcheck_interval *= -1;
 	wnd->hangcheck_timer.data = (unsigned long)wnd;
 	wnd->hangcheck_timer.function = hangcheck_proc;
-	wnd->hangcheck_timer.expires = jiffies + wnd->hangcheck_interval;
-	add_timer(&wnd->hangcheck_timer);
-	return;
+	mod_timer(&wnd->hangcheck_timer, jiffies + wnd->hangcheck_interval);
+	TRACEEXIT2(return);
 }
 
 void hangcheck_del(struct wrap_ndis_device *wnd)
@@ -1148,9 +1191,11 @@ void hangcheck_del(struct wrap_ndis_device *wnd)
 }
 
 /* worker procedure to take care of setting/checking various states */
-static void wrap_ndis_worker(void *param)
+static void wrap_ndis_worker(worker_param_t param)
 {
-	struct wrap_ndis_device *wnd = (struct wrap_ndis_device *)param;
+	struct wrap_ndis_device *wnd;
+
+	wnd = worker_param_data(param, struct wrap_ndis_device, wrap_ndis_work);
 
 	DBGTRACE2("%lu", wnd->wrap_ndis_pending_work);
 
@@ -1176,7 +1221,7 @@ NDIS_STATUS ndis_reinit(struct wrap_ndis_device *wnd)
 {
 	NDIS_STATUS status;
 
-	wnd->pm_capa = FALSE;
+	wnd->attributes &= ~NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND;
 	status = miniport_set_power_state(wnd, NdisDeviceStateD3);
 	if (status != NDIS_STATUS_SUCCESS) {
 		ERROR("halting device %s failed: %08X", wnd->net_dev->name,
@@ -1371,7 +1416,7 @@ wstdcall NTSTATUS NdisDispatchPower(struct device_object *fdo, struct irp *irp)
 		}
 		break;
 	case IRP_MN_QUERY_POWER:
-		if (wnd->pm_capa) {
+		if (wnd->attributes & NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND) {
 			ndis_status =
 				miniport_query_info(wnd, OID_PNP_QUERY_POWER,
 						    &state, sizeof(state));
@@ -1574,6 +1619,8 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	char *buf;
 	const int buf_size = 256;
 	mac_address mac;
+	struct transport_header_offset transport_header_offset;
+	int n;
 
 	ndis_status = miniport_init(wnd);
 	if (ndis_status == NDIS_STATUS_NOT_RECOGNIZED)
@@ -1583,6 +1630,13 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	wd = wnd->wd;
 	net_dev = wnd->net_dev;
 
+	ndis_status = miniport_query_info(wnd, OID_802_3_CURRENT_ADDRESS,
+					  mac, sizeof(mac));
+	if (ndis_status) {
+		ERROR("couldn't get mac address: %08X", ndis_status);
+		goto err_start;
+	}
+	DBGTRACE1("mac:" MACSTRSEP, MAC2STR(mac));
 	ndis_status = miniport_query_int(wnd, OID_GEN_PHYSICAL_MEDIUM,
 					 &wnd->physical_medium);
 	if (ndis_status != NDIS_STATUS_SUCCESS)
@@ -1591,22 +1645,15 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	get_supported_oids(wnd);
 	strncpy(net_dev->name, if_name, IFNAMSIZ - 1);
 	net_dev->name[IFNAMSIZ - 1] = '\0';
-
-	ndis_status = miniport_query_info(wnd, OID_802_3_CURRENT_ADDRESS,
-					  mac, sizeof(mac));
-	if (ndis_status) {
-		ERROR("couldn't get mac address: %08X", ndis_status);
-		return ndis_status;
-	}
-	DBGTRACE1("mac:" MACSTRSEP, MAC2STR(mac));
 	memcpy(&net_dev->dev_addr, mac, ETH_ALEN);
 
 	wnd->packet_filter = NDIS_PACKET_TYPE_DIRECTED |
-		NDIS_PACKET_TYPE_BROADCAST | NDIS_PACKET_TYPE_ALL_FUNCTIONAL;
+		NDIS_PACKET_TYPE_BROADCAST | NDIS_PACKET_TYPE_MULTICAST;
 	net_dev->open = ndis_net_dev_open;
 	net_dev->hard_start_xmit = tx_skbuff;
 	net_dev->stop = ndis_net_dev_close;
 	net_dev->get_stats = ndis_get_stats;
+	net_dev->change_mtu = ndis_change_mtu;
 	net_dev->do_ioctl = NULL;
 	if (wnd->physical_medium == NdisPhysicalMediumWirelessLan) {
 #if WIRELESS_EXT < 19
@@ -1653,13 +1700,21 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	memset(buf, 0, buf_size);
 	ndis_status = miniport_query_info(wnd, OID_GEN_VENDOR_DESCRIPTION,
 					  buf, buf_size);
-	if (ndis_status == NDIS_STATUS_SUCCESS)
-		printk(KERN_INFO "%s: vendor: '%s'\n", net_dev->name, buf);
+	if (ndis_status != NDIS_STATUS_SUCCESS) {
+		WARNING("couldn't get vendor information: 0x%x", ndis_status);
+		buf[0] = 0;
+	}
+	wnd->drv_ndis_version = n = 0;
+	miniport_query_int(wnd, OID_GEN_DRIVER_VERSION,
+			   &wnd->drv_ndis_version);
+	miniport_query_int(wnd, OID_GEN_VENDOR_DRIVER_VERSION, &n);
 
 	printk(KERN_INFO "%s: ethernet device " MACSTRSEP " using %sNDIS "
-	       "driver %s, %s\n", net_dev->name, MAC2STR(net_dev->dev_addr),
+	       "driver: %s, version: 0x%x, NDIS version: 0x%x, vendor: '%s', "
+	       "%s\n", net_dev->name, MAC2STR(net_dev->dev_addr),
 	       deserialized_driver(wnd) ? "" : "serialized ",
-	       wnd->wd->driver->name, wnd->wd->conf_file_name);
+	       wnd->wd->driver->name, n, wnd->drv_ndis_version, buf,
+	       wnd->wd->conf_file_name);
 
 	if (deserialized_driver(wnd)) {
 		/* deserialized drivers don't have a limit, but we
@@ -1691,6 +1746,22 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 		goto buffer_pool_err;
 	}
 	DBGTRACE1("pool: %p", wnd->tx_buffer_pool);
+
+	if (miniport_query_int(wnd, OID_GEN_MAXIMUM_TOTAL_SIZE, &n) ==
+	    NDIS_STATUS_SUCCESS && n > ETH_HLEN)
+		ndis_change_mtu(wnd->net_dev, n - ETH_HLEN);
+
+	if (miniport_query_int(wnd, OID_GEN_MAC_OPTIONS, &n) ==
+	    NDIS_STATUS_SUCCESS && n > 0)
+		DBGTRACE2("mac options supported: 0x%x", n);
+
+	transport_header_offset.protocol_type = NDIS_PROTOCOL_ID_TCP_IP;
+	transport_header_offset.header_offset = sizeof(ETH_HLEN);
+	ndis_status = miniport_set_info(wnd, OID_GEN_TRANSPORT_HEADER_OFFSET,
+					&transport_header_offset,
+					sizeof(transport_header_offset));
+	TRACEENTER2("%08X", ndis_status);
+
 	if (wnd->physical_medium == NdisPhysicalMediumWirelessLan) {
 		miniport_set_int(wnd, OID_802_11_POWER_MODE, NDIS_POWER_OFF);
 		miniport_set_int(wnd, OID_802_11_NETWORK_TYPE_IN_USE,
@@ -1725,6 +1796,7 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	}
 	kfree(buf);
 	wrap_procfs_add_ndis_device(wnd);
+	hangcheck_add(wnd);
 	add_stats_timer(wnd);
 	TRACEEXIT1(return NDIS_STATUS_SUCCESS);
 
@@ -1746,6 +1818,8 @@ static int ndis_remove_device(struct wrap_ndis_device *wnd)
 {
 	int tx_pending;
 
+	/* prevent setting essid during disassociation */
+	memset(&wnd->essid, 0, sizeof(wnd->essid));
 	if (wnd->physical_medium == NdisPhysicalMediumWirelessLan)
 		miniport_set_info(wnd, OID_802_11_DISASSOCIATE, NULL, 0);
 	set_bit(SHUTDOWN, &wnd->wrap_ndis_pending_work);
@@ -1753,7 +1827,6 @@ static int ndis_remove_device(struct wrap_ndis_device *wnd)
 	wnd->tx_ok = 0;
 	unregister_netdev(wnd->net_dev);
 	netif_carrier_off(wnd->net_dev);
-	cancel_delayed_work(&wnd->wrap_ndis_work);
 	/* if device is suspended, but resume failed, tx_ring_mutex
 	 * may already be locked */
 	down_trylock(&wnd->tx_ring_mutex);
@@ -1792,7 +1865,7 @@ static int ndis_remove_device(struct wrap_ndis_device *wnd)
 }
 
 static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
-				      struct device_object *pdo)
+				       struct device_object *pdo)
 {
 	struct device_object *fdo;
 	struct ndis_miniport_block *nmb;
